@@ -28,6 +28,10 @@ class Mastodon(Handler):
             self.params = params
             self.secrets = secrets
 
+        def to_tuple(self):
+            """ Convert this to a JSON-serializable tuple """
+            return self.instance, self.params, self.secrets
+
     @property
     def service_name(self):
         return "Mastodon"
@@ -47,7 +51,7 @@ class Mastodon(Handler):
     def cb_id(self):
         return 'md'
 
-    def __init__(self, name: str, token_store: dict, homepage: str = None):
+    def __init__(self, name: str, token_store: dict, timeout: int = None, homepage: str = None):
         """ Instantiate a Mastodon handler.
 
         :param str name: Human-readable website name
@@ -58,7 +62,8 @@ class Mastodon(Handler):
         """
         self._name = name
         self._homepage = homepage
-        self._pending = token_store
+        self._token_store = token_store
+        self._timeout = timeout or 600
 
     @staticmethod
     @functools.lru_cache(128)
@@ -161,7 +166,6 @@ class Mastodon(Handler):
         return disposition.Verified(id_url, redir, response)
 
     def initiate_auth(self, id_url, callback_uri, redir):
-        state = utils.gen_token()
         try:
             client = self._get_client(id_url, callback_uri)
         except Exception as err:  # pylint:disable=broad-except
@@ -170,7 +174,7 @@ class Mastodon(Handler):
         if not client:
             return disposition.Error("Failed to register OAuth client", redir)
 
-        self._pending[state] = (client, redir)
+        state = self._token_store.dumps((client.to_tuple(), redir))
 
         url = client.auth_endpoint + '?' + urllib.parse.urlencode(
             {**client.params,
@@ -183,31 +187,33 @@ class Mastodon(Handler):
         state = get.get('state')
         if not state:
             return disposition.Error("No transaction ID provided", None)
-        if state not in self._pending:
-            return disposition.Error('Transaction invalid or expired', None)
-        client, redir = self._pending[state]
 
-        if 'code' not in get:
-            return disposition.Error("Missing auth code", redir)
+        try:
+            client_tuple, redir = utils.unpack_token(self._token_store, state, self._timeout)
+        except disposition.Disposition as disp:
+            return disp
+        client = Mastodon.Client(*client_tuple)
 
-        request = requests.post(client.token_endpoint,
-                                {**client.params,
-                                 **client.secrets,
-                                 'grant_type': 'authorization_code',
-                                 'code': get['code']})
-        if request.status_code != 200:
-            LOGGER.warning('oauth/token: %d %s', request.status_code, request.text)
-            return disposition.Error("Could not retrieve access token", redir)
+        if 'error' in get:
+            return disposition.Error("Error signing into Mastodon", redir)
 
-        response = request.json()
-        if 'access_token' not in response:
-            LOGGER.warning("Response did not contain 'access_token': %s", response)
-            return disposition.Error("No access token provided", redir)
+        try:
+            request = requests.post(client.token_endpoint,
+                                    {**client.params,
+                                     **client.secrets,
+                                     'grant_type': 'authorization_code',
+                                     'code': get['code']})
+            if request.status_code != 200:
+                LOGGER.warning('oauth/token: %d %s', request.status_code, request.text)
+                return disposition.Error("Could not retrieve access token", redir)
 
-        token = response['access_token']
-        auth_headers = {'Authorization': 'Bearer ' + token}
+            response = request.json()
+            token = response['access_token']
+            auth_headers = {'Authorization': 'Bearer ' + token}
 
-        result = self._get_identity(client, auth_headers, redir)
+            result = self._get_identity(client, auth_headers, redir)
+        except KeyError as key:
+            result = disposition.Error("Missing " + str(key), redir)
 
         # try to clean up after ourselves
         request = requests.post(client.revoke_endpoint, data={
@@ -231,6 +237,9 @@ def from_config(config, token_store):
 
     MASTODON_NAME -- the name of your website (required)
     MASTODON_HOMEPAGE -- your website's homepage (recommended)
+    MASTODON_TIMEOUT -- the maximum time to wait for login to complete
     """
 
-    return Mastodon(config['MASTODON_NAME'], token_store, config.get('MASTODON_HOMEPAGE'))
+    return Mastodon(config['MASTODON_NAME'], token_store,
+                    timeout=config.get('MASTODON_TIMEOUT'),
+                    homepage=config.get('MASTODON_HOMEPAGE'))
